"use strict";
var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all2) => {
  for (var name in all2)
    __defProp(target, name, { get: all2[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  // If the importer is in node compatibility mode or this is not an ESM
  // file that has been converted to a CommonJS file using a Babel-
  // compatible transform (i.e. "__esModule" has not been set), then set
  // "default" to the CommonJS "module.exports" for node compatibility.
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  Coroutine: () => Coroutine,
  CoroutineController: () => CoroutineController,
  CoroutineManager: () => CoroutineManager,
  FocusTargetController: () => FocusTargetController,
  FocusTargetManager: () => FocusTargetManager,
  waitUntil: () => awaiters_exports
});
module.exports = __toCommonJS(src_exports);

// src/focus/constants.ts
var FOCUS_TARGET_OPTIONS_KEY = Symbol(
  "focus target options"
);

// src/focus/manager/FocusTargetManager.ts
var FocusTargetManager = class {
  #controllerInterface;
  constructor(controllerInterface) {
    this.#controllerInterface = controllerInterface;
  }
  createFocusTarget(options = {}) {
    const target = {
      [FOCUS_TARGET_OPTIONS_KEY]: options,
      focus: () => this.#controllerInterface.focus(target),
      blur: () => this.#controllerInterface.blur(target)
    };
    return target;
  }
  createCombinedFocusTarget(sources) {
    const target = {
      [FOCUS_TARGET_OPTIONS_KEY]: {},
      focus: () => this.#throwCombinedFocusError(),
      blur: () => this.#controllerInterface.blur(target)
    };
    this.#controllerInterface.registerCombinedFocusTarget(target, sources);
    return target;
  }
  #throwCombinedFocusError() {
    throw new Error("Cannot focus a combined focus target");
  }
};

// src/focus/manager/FocusTargetController.ts
var FocusTargetController = class {
  #combinedFocusTargets = /* @__PURE__ */ new WeakMap();
  #currentFocusTarget = null;
  createManager() {
    const controllerInterface = this.#createControllerManagerInterface();
    return new FocusTargetManager(controllerInterface);
  }
  /**
   * Checks if the provided focus target is currently active.
   */
  isFocused(target) {
    const combined = this.#combinedFocusTargets.get(target);
    if (combined) {
      return combined.some((t) => this.isFocused(t));
    } else {
      const options = target[FOCUS_TARGET_OPTIONS_KEY];
      if (!options.require && this.#currentFocusTarget === null)
        return true;
      return this.#currentFocusTarget === target;
    }
  }
  isAnyFocusTargetActive() {
    return this.#currentFocusTarget !== null;
  }
  /**
   * Returns `null` if there is no active focus target,
   * `undefined` if there is but it has no name set,
   * or the name if it's set.
   */
  getActiveFocusTargetDisplayName() {
    if (this.#currentFocusTarget === null) {
      return null;
    } else {
      const options = this.#currentFocusTarget[FOCUS_TARGET_OPTIONS_KEY];
      return options.displayName;
    }
  }
  #createControllerManagerInterface() {
    return {
      focus: this.#focus.bind(this),
      blur: this.#blur.bind(this),
      registerCombinedFocusTarget: this.#registerCombinedFocusTarget.bind(this)
    };
  }
  #focus(target) {
    this.#currentFocusTarget = target;
  }
  #blur(target) {
    if (this.#currentFocusTarget === target) {
      this.#currentFocusTarget = null;
    }
  }
  #registerCombinedFocusTarget(target, sources) {
    this.#combinedFocusTargets.set(target, sources);
  }
};

// src/manager/Coroutine.ts
var import_tiny_invariant = __toESM(require("tiny-invariant"));

// src/manager/awaiters/nextTick.ts
function nextTick(options = {}) {
  return {
    ...options,
    identifier: "nextTick",
    shouldContinue() {
      return { done: true };
    }
  };
}

// src/manager/NestedCoroutineHandler.ts
var NestedCoroutineHandler = class {
  #options;
  #castedAwaiters;
  #checkCount = 0;
  constructor(options) {
    this.#options = options;
    this.#castedAwaiters = options.awaiters.map((awaiter) => {
      if (typeof awaiter === "function") {
        return new Coroutine({
          identifier: awaiter.name,
          coroutine: awaiter
        });
      } else if (!awaiter) {
        return nextTick();
      } else {
        return awaiter;
      }
    });
    const stackTrace = [...this.#options.parentStackTrace];
    if (this.#options.identifier)
      stackTrace.push(this.#options.identifier);
    for (const handler of this.#castedAwaiters) {
      if (!(handler instanceof Coroutine))
        continue;
      handler.start({
        delete: () => this.#deleteHandler(handler),
        isFocused: options.isFocused,
        parentStackTrace: stackTrace
      });
    }
  }
  dispose() {
    for (const handler of this.#castedAwaiters) {
      if (!(handler instanceof Coroutine))
        continue;
      handler.dispose();
    }
  }
  handle(ctx) {
    const results = [];
    for (const handler of this.#castedAwaiters) {
      if (handler instanceof Coroutine) {
        handler.handle(ctx);
        results.push(null);
        this.#checkCount += handler.getLastCheckCount();
      } else {
        const res = handler.shouldContinue(ctx);
        results.push(res);
        this.#checkCount++;
      }
    }
    const fullResults = results.map(
      (result, index) => {
        if (result === null) {
          return {
            done: this.#castedAwaiters[index] === null
          };
        } else
          return result;
      }
    );
    return this.#options.handler(ctx, fullResults);
  }
  /**
   * Returns the number of `shouldContinue` calls that ran on the last `handle` call.
   *
   * Useful for displaying debugging information.
   */
  getLastCheckCount() {
    return 0;
  }
  #deleteHandler(handler) {
    const index = this.#castedAwaiters.indexOf(handler);
    if (index === -1) {
      throw new Error(
        "Handler does not exist, it may have been deleted already"
      );
    }
    this.#castedAwaiters[index] = null;
  }
};

// src/manager/awaiters/exotic/dispose.ts
var marker = Symbol("dispose hook");
function createDisposeHook(callback) {
  return {
    marker,
    callback
  };
}
function isDisposeHook(value) {
  return value?.marker === marker;
}

// src/manager/awaiters/exotic/isExoticAwaiter.ts
function isExoticAwaiter(awaiter) {
  if (!awaiter)
    return false;
  if (typeof awaiter !== "object")
    return false;
  const casted = awaiter;
  return typeof casted.marker === "symbol";
}

// src/manager/awaiters/exotic/nest.ts
var marker2 = Symbol("nest");
function nest(options) {
  const awaiter = {
    marker: marker2,
    options
  };
  return awaiter;
}
function isNestAwaiter(value) {
  return value?.marker === marker2;
}

// src/manager/awaiters/exotic/options.ts
var marker3 = Symbol("options hook");
function createOptionsHook(options) {
  return {
    marker: marker3,
    options
  };
}
function isOptionsHook(value) {
  return value?.marker === marker3;
}

// src/manager/Coroutine.ts
var ERROR_AWAITER_SYMBOL = Symbol();
var Coroutine = class {
  #identifier;
  #coroutineFn;
  #disposeHandlers = /* @__PURE__ */ new Set();
  #coroutineOrNull = null;
  #startInterfaceOrNull = null;
  #stackTraceOrNull = null;
  /**
   * Whether the coroutine has been disposed.
   */
  #isDisposed = false;
  /**
   * An error holding the stack trace for the first time this coroutine was disposed.
   * Helpful for debugging why a coroutine was disposed twice.
   */
  #firstDisposalTrace = null;
  /**
   * Whether the coroutine is waiting for a {@link CoroutineAwait.delay} delay.
   */
  #waitingForDelay = false;
  /**
   * The awaiter returned by the last `yield` call.
   */
  #currentAwaiter = null;
  #currentNest = null;
  /**
   * Awaiter options used as the base.
   */
  #baseAwaiterOptions = {};
  #checkCount = 0;
  constructor({ identifier, coroutine }) {
    this.#identifier = identifier;
    this.#coroutineFn = coroutine;
  }
  get #startInterface() {
    (0, import_tiny_invariant.default)(
      this.#startInterfaceOrNull !== null,
      "Coroutine has not been started yet"
    );
    return this.#startInterfaceOrNull;
  }
  get #coroutine() {
    (0, import_tiny_invariant.default)(
      this.#coroutineOrNull !== null,
      "Coroutine has not been started yet"
    );
    return this.#coroutineOrNull;
  }
  get #stackTrace() {
    (0, import_tiny_invariant.default)(
      this.#stackTraceOrNull !== null,
      "Coroutine has not been started yet"
    );
    return this.#stackTraceOrNull;
  }
  start(startInterface) {
    this.#startInterfaceOrNull = startInterface;
    this.#coroutineOrNull = this.#coroutineFn();
    this.#stackTraceOrNull = [
      ...startInterface.parentStackTrace,
      this.#identifier
    ];
  }
  dispose() {
    if (this.#isDisposed) {
      throw new Error("Coroutine has already been disposed", {
        cause: this.#firstDisposalTrace
      });
    }
    for (const disposeHandler of this.#disposeHandlers) {
      disposeHandler();
    }
    this.#isDisposed = true;
    this.#currentNest?.dispose();
    this.#startInterface.delete();
    this.#firstDisposalTrace = new Error("Coroutine disposed here");
  }
  handle(ctx) {
    if (this.#isDisposed) {
      throw new Error(
        "Cannot handle a coroutine after it has been disposed",
        {
          cause: this.#firstDisposalTrace
        }
      );
    }
    this.#checkCount = 0;
    try {
      if (this.#isDelaying())
        return;
      if (this.#isOutOfFocus())
        return;
      if (this.#handleNest(ctx))
        return;
      if (this.#handleCurrentAwaiter(ctx))
        return;
      this.#loadNextAwaiter(ctx, {
        done: true,
        data: void 0
      });
    } catch (err) {
      throw this.#addTraceToError(err);
    }
  }
  /**
   * Returns the identifier of this coroutine.
   *
   * Useful for displaying debug information.
   */
  getIdentifier() {
    return this.#identifier;
  }
  /**
   * Returns the number of `shouldContinue` calls that ran on the last `handle` call.
   *
   * Useful for displaying debugging information.
   */
  getLastCheckCount() {
    return 0;
  }
  #isDelaying() {
    if (this.#waitingForDelay)
      return true;
    const delayPromise = this.#currentAwaiter?.delay;
    if (!delayPromise)
      return false;
    this.#waitingForDelay = true;
    delayPromise.then(() => {
      this.#waitingForDelay = false;
    });
    return true;
  }
  #isOutOfFocus() {
    const focusTarget = this.#currentAwaiter?.focusTarget;
    if (!focusTarget)
      return false;
    return !this.#startInterface.isFocused(focusTarget);
  }
  #handleNest(ctx) {
    if (!this.#currentNest)
      return false;
    const awaiterResult = this.#currentNest.handle(ctx);
    this.#checkCount += this.#currentNest.getLastCheckCount();
    if (awaiterResult.done) {
      this.#currentNest = null;
      this.#loadNextAwaiter(ctx, awaiterResult.data);
    }
    return true;
  }
  #loadNextAwaiter(ctx, awaiterResult) {
    let nextAwaiter = this.#getNextAwaiter(ctx, awaiterResult);
    let nextAwaiterValue = nextAwaiter.value && this.#castAwaiter(nextAwaiter.value);
    while (nextAwaiterValue && isExoticAwaiter(nextAwaiterValue)) {
      this.#handleExoticAwaiter(nextAwaiterValue);
      if (this.#currentNest)
        return;
      nextAwaiter = this.#getNextAwaiter(ctx, awaiterResult);
      nextAwaiterValue = this.#castAwaiter(nextAwaiter.value);
    }
    (0, import_tiny_invariant.default)(
      !isExoticAwaiter(nextAwaiterValue),
      "Coroutine hook was not handled"
    );
    if (nextAwaiter.done) {
      this.dispose();
    } else {
      this.#currentAwaiter = nextAwaiterValue;
    }
  }
  #getNextAwaiter(ctx, awaiterResult) {
    return this.#coroutine.next(
      this.#createYieldResult(ctx, awaiterResult)
    );
  }
  #createYieldResult(ctx, data) {
    return {
      ctx,
      data
    };
  }
  #handleExoticAwaiter(hook) {
    if (isDisposeHook(hook)) {
      this.#disposeHandlers.add(hook.callback);
    } else if (isOptionsHook(hook)) {
      Object.assign(this.#baseAwaiterOptions, hook.options);
    } else if (isNestAwaiter(hook)) {
      this.#startNestedCoroutine(hook.options);
    } else {
      const markerDescription = hook.marker?.description;
      throw new Error(
        `Invalid hook marker: ${markerDescription ?? JSON.stringify(hook)}`
      );
    }
  }
  #castAwaiter(awaiter) {
    if (!awaiter) {
      return nextTick();
    } else if (awaiter instanceof Coroutine || typeof awaiter === "function") {
      return nest({
        identifier: null,
        handler(_ctx, results) {
          const result = results[0];
          (0, import_tiny_invariant.default)(result, "Nest handler returned no result");
          return result;
        },
        awaiters: [awaiter]
      });
    } else {
      return awaiter;
    }
  }
  #startNestedCoroutine(options) {
    this.#currentNest = new NestedCoroutineHandler({
      ...options,
      isFocused: this.#startInterface.isFocused,
      parentStackTrace: this.#stackTrace
    });
  }
  #handleCurrentAwaiter(ctx) {
    if (!this.#currentAwaiter)
      return false;
    try {
      const result = this.#currentAwaiter.shouldContinue(ctx);
      this.#checkCount++;
      if (result.done) {
        this.#loadNextAwaiter(ctx, result);
      }
      return true;
    } catch (err) {
      if (!(err instanceof Error))
        throw err;
      err[ERROR_AWAITER_SYMBOL] = this.#currentAwaiter.identifier;
      throw err;
    }
  }
  #addTraceToError(err) {
    if (!(err instanceof Error))
      return err;
    const errorCasted = err;
    const trace = [...this.#stackTrace];
    if (errorCasted[ERROR_AWAITER_SYMBOL]) {
      trace.push(errorCasted[ERROR_AWAITER_SYMBOL]);
      delete errorCasted[ERROR_AWAITER_SYMBOL];
    }
    const traceText = trace.map((line) => `  at ${line}`).join("\n");
    err.message += `

${traceText}`;
    return err;
  }
};

// src/manager/CoroutineManager.ts
var CoroutineManager = class {
  #controllerInterface;
  #focusTargetController = new FocusTargetController();
  #focusTargetManager = this.#focusTargetController.createManager();
  constructor(controllerInterface) {
    this.#controllerInterface = controllerInterface;
  }
  startCoroutine(coroutine) {
    if (coroutine instanceof Coroutine) {
      coroutine.start({
        isFocused: (target) => this.#focusTargetController.isFocused(target),
        delete: () => this.#controllerInterface.deregisterCoroutine(coroutine),
        parentStackTrace: []
      });
      this.#controllerInterface.registerCoroutine(coroutine);
      return this.#createCoroutineResult(coroutine);
    } else {
      return this.startCoroutine(
        new Coroutine({
          identifier: coroutine.name,
          coroutine
        })
      );
    }
  }
  getFocusTargetManager() {
    return this.#focusTargetManager;
  }
  /**
   * When yielded, registers the callback to be called when this coroutine is about to be disposed.
   * This is always during a `yield` statement (which will never complete).
   */
  hookDispose(callback) {
    return createDisposeHook(callback);
  }
  /**
   * When yielded, sets the default awaiter options for a coroutine, for any calls after this one.
   * Calling multiple times shallow merges the options.
   */
  hookOptions(options) {
    return createOptionsHook(options);
  }
  isFocusGlobal() {
    return !this.#focusTargetController.isAnyFocusTargetActive();
  }
  #createCoroutineResult(coroutine) {
    return {
      stop: () => coroutine.dispose()
    };
  }
};

// src/manager/CoroutineController.ts
var CoroutineController = class {
  #manager = new CoroutineManager(
    this.#createControllerManagerInterface()
  );
  #coroutines = /* @__PURE__ */ new Set();
  #disposedCountThisTick = 0;
  #activeFocusTarget = null;
  getManager() {
    return this.#manager;
  }
  /**
   * Runs all coroutines.
   * @param ctx Context returned as `.ctx` from any `yield`s.
   */
  tick(ctx) {
    this.#disposedCountThisTick = 0;
    for (const coroutine of this.#coroutines) {
      coroutine.handle(ctx);
    }
  }
  /**
   * Returns the number of coroutines currently registered.
   *
   * Useful for showing debug information.
   */
  getCoroutineCount() {
    return this.#coroutines.size;
  }
  /**
   * Returns the number of `shouldContinue` calls that ran in the last `tick()` call.
   *
   * Useful for showing debug information.
   */
  getLastCheckCount() {
    let count = 0;
    for (const coroutine of this.#coroutines) {
      count += coroutine.getLastCheckCount();
    }
    return count;
  }
  /**
   * Returns the number of coroutines that were disposed in the last `tick()` call.
   *
   * Useful for showing debug information.
   */
  getDisposedCountThisTick() {
    return this.#disposedCountThisTick;
  }
  getRegisteredIdentifiers() {
    return Array.from(this.#coroutines).map((c) => c.getIdentifier());
  }
  getActiveFocusTargetIdentifier() {
    return this.#activeFocusTarget;
  }
  #createControllerManagerInterface() {
    return {
      registerCoroutine: (coroutine) => this.#registerCoroutine(coroutine),
      deregisterCoroutine: (coroutine) => this.#deregisterCoroutine(coroutine),
      handleActiveFocusTargetChanged: (identifier) => this.#activeFocusTarget = identifier
    };
  }
  #registerCoroutine(coroutine) {
    this.#coroutines.add(coroutine);
  }
  #deregisterCoroutine(coroutine) {
    if (!this.#coroutines.has(coroutine)) {
      throw new Error(
        "Coroutine was never registered or has already been deregistered"
      );
    }
    this.#coroutines.delete(coroutine);
    this.#disposedCountThisTick++;
  }
};

// src/manager/awaiters/index.ts
var awaiters_exports = {};
__export(awaiters_exports, {
  all: () => all,
  delay: () => delay,
  nest: () => nest,
  nextTick: () => nextTick,
  one: () => one
});

// src/manager/awaiters/one.ts
var import_tiny_invariant2 = __toESM(require("tiny-invariant"));
function one(awaiters) {
  return nest({
    identifier: "waitUntil.one",
    awaiters,
    handler(_ctx, results) {
      for (let i = 0; i < results.length; i++) {
        const result = results[i];
        (0, import_tiny_invariant2.default)(result, "Missing item in results array");
        if (result.done) {
          return {
            done: true,
            data: i
          };
        }
      }
      return { done: false };
    }
  });
}

// src/manager/awaiters/all.ts
var import_tiny_invariant3 = __toESM(require("tiny-invariant"));
function all(awaiters) {
  return nest({
    identifier: "waitUntil.one",
    awaiters,
    handler(_ctx, results) {
      let allDone = true;
      for (let i = 0; i < results.length; i++) {
        const result = results[i];
        (0, import_tiny_invariant3.default)(result, "Missing item in results array");
        if (!result.done) {
          allDone = false;
          break;
        }
      }
      return { done: allDone, data: void 0 };
    }
  });
}

// src/manager/awaiters/delay.ts
function delay(ms, options = {}) {
  let done = false;
  const delay2 = new Promise((yay) => {
    setTimeout(() => {
      done = true;
      yay();
    }, ms);
  });
  return {
    ...options,
    identifier: "waitUntil.delay",
    delay: delay2,
    shouldContinue() {
      return { done };
    }
  };
}
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  Coroutine,
  CoroutineController,
  CoroutineManager,
  FocusTargetController,
  FocusTargetManager,
  waitUntil
});
